package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.signature.Signature
import kotlinx.datetime.Instant
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.utools.now

/**
 * Secret key to sign only
 */
@Serializable
class SigningSecretKey(
        override val keyBytes: UByteArray,
        @Transient
        private var cachedPublicKey: SigningPublicKey?=null
    ) : BinaryKeyBase(), SigningKey {

    override val verifyingKey: SigningPublicKey by lazy {
        cachedPublicKey ?:
            SigningPublicKey(Signature.ed25519SkToPk(keyBytes)).also { cachedPublicKey = it }
    }

    override fun sign(message: UByteArray): UByteArray = Signature.detached(message, keyBytes)

    override fun seal(message: UByteArray, expiresAt: Instant?): Seal =
        Seal.create(this, message, now(), expiresAt)

    override fun toString(): String = "Sct:${super.toString()}"

    override val id: KeyId = verifyingKey.id

    @Transient
    override val magick = KeysMagickNumber.defaultSigning

    companion object {

        data class SigningKeyPair(val secretKey: SigningSecretKey, val publicKey: SigningPublicKey)

        fun generatePair(): SigningKeyPair {
            val p = Signature.keypair()
            val publicKey = SigningPublicKey(p.publicKey)
            return SigningKeyPair(SigningSecretKey(p.secretKey, publicKey), publicKey)
        }

        fun new(): SigningSecretKey = generatePair().secretKey

    }
}