package net.sergeych.crypto2

import kotlinx.serialization.Serializable

@Serializable
class UniversalRing(
    val keyWithTags: Map<UniversalKey,Set<String>>
) {
    constructor(vararg keys: UniversalKey) : this(keys.associateWith { setOf() })
    constructor(vararg keys: DecryptingKey) : this(keys.associate { UniversalKey.from(it) to setOf<String>() })
    constructor(vararg keyTags: Pair<UniversalKey, String>)
            : this(keyTags.associate { it.first to setOf(it.second) })

    val decryptingKeys: Set<DecryptingKey> by lazy { keys<DecryptingKey>() }

    inline fun <reified T>keys(): Set<T> =
        keyWithTags.keys.mapNotNull { it as? T }.toSet()

    inline fun <reified T>findKey(id: KeyId): UniversalKey? =
        keyWithTags.keys.find { it is T && it.id == id }

    fun allByAnyOfTags(vararg tags: String) = sequence {
        for( e in keyWithTags.entries) {
            if( tags.any { it in e.value }) yield(e.key)
        }
    }

    inline fun <reified T>keyByTag(tag: String) = allByAnyOfTags(tag).first { it is T }

    @Suppress("unused")
    inline fun <reified T>keyByAnyTag(vararg tags: String) = allByAnyOfTags(*tags).first { it is T }

    operator fun get(keyId: KeyId): Collection<UniversalKey> = keyWithTags.keys.filter { it.id == keyId }

    fun getTags(key: UniversalKey): Set<String>? = keyWithTags[key]

    operator fun contains(element: UniversalKey): Boolean = element in keyWithTags

    operator fun plus(key: UniversalKey): UniversalRing =
        if( key in this ) this else UniversalRing(keyWithTags + (key to setOf()) )

    operator fun plus(keyTag: Pair<UniversalKey,String>): UniversalRing =
        if( keyTag.first in this )
            addTags(keyTag.first,keyTag.second)
        else UniversalRing(keyWithTags + keyWithTags)

    fun add(key: UniversalKey, vararg tags: String): UniversalRing =
        UniversalRing(keyWithTags + (key to tags.toSet()) )

    fun add(key: UniversalKey, tags: Collection<String>): UniversalRing =
        UniversalRing(keyWithTags + (key to tags.toSet()) )

    fun addTags(key: UniversalKey,tags: Set<String>): UniversalRing {
        val kt1 = keyWithTags.toMutableMap()
        kt1[key]?.let {
            kt1[key] = it + tags
        } ?: run {
            kt1[key] = tags
        }
        return UniversalRing(kt1)
    }

    fun addTags(key: UniversalKey,vararg tags: String): UniversalRing =
        addTags(key,tags.toSet())

    operator fun minus(key: UniversalKey): UniversalRing =
        if( key in this ) UniversalRing(keyWithTags.filter { it.key != key }) else this

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other !is UniversalRing) return false
        return keyWithTags == other.keyWithTags
    }

    override fun toString(): String {
        val ss = keyWithTags.entries
            .joinToString(","){"${it.value.joinToString{ ":" }}:${it.key}"}
        return "Kr[$ss]"
    }

    override fun hashCode(): Int {
        return keyWithTags.hashCode()
    }

    infix fun equalKeys(other: UniversalRing): Boolean = keyWithTags.keys == other.keyWithTags.keys

    fun removeTags(key: UniversalKey, vararg tags: String): UniversalRing = removeTags(key, tags.toSet())

    fun removeTags(key: UniversalKey, tags: Set<String>): UniversalRing {
        val kt1 = keyWithTags.toMutableMap()
        kt1[key]?.let {
            kt1[key] = it - tags
        }
        return UniversalRing(kt1)
    }


}
